//
//  UserPostCommentManager.swift
//  SampleApp
//
//  Created by Michael Abadi on 10/06/20.
//  Copyright © 2020 David Zhang. All rights reserved.
//


import Foundation
import AmitySDK
import UIKit

typealias DefaultCompletion = ((Bool, Error?) -> Void)?

struct UserPostCommentModel {
    let commentId: String
    let displayName: String
    let text: String
    let date: String
    let reaction: String
    let isEdited: Bool
    let isDeleted: Bool
    let metadata: [String: Any]?
    let mentionees: [AmityMentionees]?
    let isParent: Bool
    let comment: AmityComment
}

enum UserPostCommentItem {
    case parent(UserPostCommentModel)
    case child(UserPostCommentModel)
}

class UserPostCommentManager {
    
    let client: AmityClient
    let commentRepository: AmityCommentRepository
    
    var commentCollection: AmityCollection<AmityComment>?
    var commentCollectionToken: AmityNotificationToken?
    
    let reactionRepo: AmityReactionRepository
    
    var editedComment: UserPostCommentModel?
    var selectedComment: AmityComment?
    let userId: String?
    let userName: String?
    let postId: String?
    var parentCommentId: String?

    var includeDeleted = false
    var commentOrderType: AmityOrderBy = .descending
    var commentDataTypes: AmityCommentDataTypeFilter?
    
    private var items: [UserPostCommentItem] = []
    
    let communityId: String?
    
    var editor: AmityCommentEditor?
    
    var flagger: AmityCommentFlagger?
    
    // Selected files to create comment
    var selectedImages = [String]()
    
    lazy var dateFormatter: DateFormatter = {
        let formatter = DateFormatter()
        formatter.dateFormat = "MMM d, HH:mm"
        return formatter
    }()
    
    init(client: AmityClient, postId: String?, parentCommentId: String? = nil, userId: String?, userName: String? = nil, communityId: String?) {
        self.client = client
        self.userId = userId
        self.userName = userName
        self.parentCommentId = parentCommentId
        self.postId = postId
        self.commentRepository = AmityCommentRepository(client: client)
        self.reactionRepo = AmityReactionRepository(client: client)
        self.communityId = communityId
    }
    
    // MARK:- Comment Observer
    
    // Listen to comment changes. Anytime a comment is created, observer
    // would be notified.
    // Note: Retain that on AmityNotificationToken
    func observeCommentFeedChanges(changeHandler: @escaping ()->()) {
        let options = AmityCommentQueryOptions(referenceId: postId!, referenceType: .post, filterByParentId: true, dataTypes: commentDataTypes, parentId: parentCommentId, orderBy: commentOrderType, includeDeleted: includeDeleted)
        commentCollection = commentRepository.getComments(with: options)
        commentCollectionToken = commentCollection?.observe { [weak self] (collection, _, error) in
            
            Log.add(info: "[COMMENT] Change Observed in Comments \(collection.count())")
            self?.prepareData()
            changeHandler()
        }
    }
    
    func toggleReactionForComment(comment: AmityComment) {
        Task { @MainActor in
            let reactionName = "like"
            
            if comment.myReactions.contains(reactionName) {
                do {
                    let result = try await reactionRepo.removeReaction(reactionName, referenceId: comment.commentId, referenceType: .comment)
                    Log.add(info: "[SampleApp]: Comment reaction removed \(result)")
                } catch {
                    Log.add(info: "[SampleApp]: Comment reaction removed error: \(String(describing: error))")
                }
    
            } else {
                do {
                    let result = try await reactionRepo.addReaction(reactionName, referenceId: comment.commentId, referenceType: .comment)
                    Log.add(info: "[SampleApp]: Comment reaction added \(result)")
                } catch {
                    Log.add(info: "[SampleApp]: Comment reaction added error: \(String(describing: error))")
                }
            }
        }
    }
    
    // MARK: CRUD Comment

    var newComment: AmityObject<AmityComment>?
    var newCommentToken: AmityNotificationToken?
    
    func createComment(text: String, attachments: [AmityCommentAttachment], metadata: [String: Any]?, mentionees: AmityMentioneesBuilder?, completion: @escaping AmityRequestCompletion) {
        
        let options = AmityCommentCreateOptions(referenceId: postId!, referenceType: .post, text: text, attachments: attachments, parentId: parentCommentId, metadata: metadata, mentioneeBuilder: mentionees)
        Task { @MainActor in
            do {
                let comment = try await commentRepository.createComment(with: options)
                completion(true, nil)
            } catch {
                completion(false, error)
            }
        }
    }
        
    func updateComment(text: String?, attachments: [AmityCommentAttachment], metadata: [String: Any]?, mentionees: AmityMentioneesBuilder?, onCompletion: @escaping AmityRequestCompletion ) {
        guard let commentToEdit = editedComment else { return }
        
        let updateOptions = AmityCommentUpdateOptions(text: text, attachments: attachments, metadata: metadata, mentioneesBuilder: mentionees)
        
        Task { @MainActor in
            do {
                let _ = try await commentRepository.updateComment(withId: commentToEdit.commentId, options: updateOptions)
                onCompletion(true, nil)
            } catch {
                onCompletion(false, error)
            }
            editedComment = nil
        }
    }
    
    func deleteComment(at index: Int, hardDelete: Bool, onCompletion: @escaping (_ isSuccess: Bool)->()) {
        Task { @MainActor in
            do {
                var commentId: String
                
                switch items[index] {
                case .child(let replyModel):
                    commentId = replyModel.commentId
                case .parent(let commentModel):
                    commentId = commentModel.commentId
                }
                
                let success: Bool
                
                if hardDelete {
                    success = try await commentRepository.hardDeleteComment(withId: commentId)
                } else {
                    success = try await commentRepository.softDeleteComment(withId: commentId)
                }
                
                onCompletion(success)
            } catch {}
            editedComment = nil
        }
    }
    
    // MARK: Flagging Comment
    
    func flagComment(at index: Int, onCompletion: @escaping (_ isSuccess: Bool, _ error: Error?)->()) {
        Task { @MainActor in
            let comment = getCommentFromIndex(index: index)
            selectedComment = comment
            guard let commentId = selectedComment?.commentId else { return }
            
            do {
                let success = try await commentRepository.flagComment(withId: commentId)
                onCompletion(success, nil)
            } catch {
                onCompletion(false, error)
            }
            selectedComment = nil
        }
    }
    
    func unflagComment(at index: Int, onCompletion: @escaping (_ isSuccess: Bool, _ error: Error?)->()) {
        Task { @MainActor in
            let comment = getCommentFromIndex(index: index)
            selectedComment = comment
            guard let commentId = selectedComment?.commentId else { return }
            
            do {
                let success = try await commentRepository.unflagComment(withId: commentId)
                onCompletion(success, nil)
            } catch {
                onCompletion(false, error)
            }
            selectedComment = nil
        }
    }
    
    func getCommentFromIndex(index: Int) -> AmityComment {
        let item = items[index]
        switch item {
        case .child(let comment), .parent(let comment):
            return comment.comment
        }
    }
    
    // MARK: Public Helper method
    
    func loadMoreComments() {
        if commentOrderType == .descending {
            guard let hasMoreComments = commentCollection?.hasPrevious, hasMoreComments else { return }
            commentCollection?.previousPage()
        } else {
            guard let hasMoreComments = commentCollection?.hasNext, hasMoreComments else { return }
            commentCollection?.nextPage()
        }
    }
    
    func getNumberOfCommentItems() -> Int {
        return items.count
    }
    
    func getCommentItemHeaderData(at index: Int) -> (title: String, date: String) {
        switch items[index] {
        case .parent(let comment):
            return (comment.displayName, comment.date)
        case .child(let comment):
            return (comment.displayName, comment.date)
        }
    }
    
    func getCommentId(at index: Int) -> String {
        switch items[index] {
        case .parent(let comment):
            return comment.commentId
        case .child(let comment):
            return comment.commentId
        }
    }
    
    func getCommentItem(at index: Int) -> UserPostCommentItem {
        return items[index]
    }
    
    func prepareToFlagComment(comment: AmityComment) {
        self.selectedComment = comment
    }
    
    func prepareToEditComment(at index: Int) {
        switch items[index] {
        case .parent(let comment):
            self.editedComment = comment
        case .child(let comment):
            self.editedComment = comment
        }
    }
    
    func getEditCommentData() -> UserPostCommentModel? {
        return editedComment
    }
    
    func getReadableDate(date: Date?) -> String {
        if let date = date {
            return dateFormatter.string(from: date)
        } else {
            return ""
        }
    }
    
    func isFlaggedByMe(onCompletion: @escaping (_ isSuccess: Bool)->()){
        if let selectedComment = selectedComment {
            Task { @MainActor in
                do {
                    let isFlagged = try await commentRepository.isCommentFlaggedByMe(withId: selectedComment.commentId)
                    onCompletion(isFlagged)
                } catch {
                    onCompletion(false)
                }
            }
        }
    }
    
    // MARK:- Private Helpers
    
    private func prepareData() {
        guard let commentCollection = commentCollection else { return }
        
        var items: [UserPostCommentItem] = []
        for comment in commentCollection.allObjects() {
            // append itself
            items.append(.parent(comment.asPostCommentModel()))
            
            // append comment children
            let childrenComments: [UserPostCommentItem] = comment.childrenComments.map { UserPostCommentItem.child($0.asPostCommentModel()) }
            items += childrenComments
        }
        self.items = items
    }
    
}

private extension AmityComment {
    
    func asPostCommentModel() -> UserPostCommentModel {
        var textData = ""
        
        if let text = data?["text"] as? String, text != "" {
            textData += text + "\n\n"
        }
        
        let sortedDataTypes = dataTypes.map { $0.rawValue }.sorted(by: { $0 < $1 })

        textData += "Comment Id: \(self.commentId)\n"
        textData += "DataTypes: \(sortedDataTypes)\n"
        textData += "Reference Type: \(self.referenceType.description)\n"
        textData += "Reference Id: \(self.referenceId)\n"
        textData += "My Reactions: \(self.myReactions.joined(separator: ","))\n"
        textData += "Metadata: \(String(describing: self.metadata))"
        textData += "Children Number: \(self.childrenNumber)\n"
        textData += "Is Deleted: \(self.isDeleted)\n"
        switch self.target {
        case .community(_, let creatorMember):
            textData += "TargetType: \(self.target.getStringValue())\n"
            if let creatorMember {
                textData += "CommunityUserRole: \(creatorMember.roles)\n"
            } else {
                textData += "CommunityUserRole: []\n"
            }
        case .content, .user:
            textData += "TargetType: \(self.target.getStringValue())\n"
        default:
            break
        }
        
        
        let displayName = user?.displayName ?? "No Name"
        
        if !self.attachments.isEmpty {
            textData += "\n\n\(self.attachments.count) Attachments:\n"
            for attachment in self.attachments {
                switch attachment {
                case .image(let fileId, let data):
                    let fileInfo = "\n- File Id: \(fileId) URL: \(data?.fileURL ?? "-")"
                    
                    textData += "\n\(fileInfo)"
                @unknown default:
                    let fileInfo = "Unknown Attachment"
                    textData += "\n\(fileInfo)"
                }
            }
        }
        
        let formatter = DateFormatter()
        formatter.dateFormat = "MMM d, HH:mm"
        let dateStr = formatter.string(from: createdAt)
        
        var formattedReaction = ""
        
        if let raction = reactions as? [String: Int] {
            formattedReaction = raction.map { $1 > 0 ? "\($0): \($1)" : "" }.joined(separator: " ")
        }
                
        return UserPostCommentModel(commentId: commentId, displayName: displayName, text: textData, date: dateStr, reaction: formattedReaction, isEdited: isEdited, isDeleted: self.isDeleted, metadata: metadata, mentionees: mentionees, isParent: self.parentId == nil, comment: self)
    }
    
}
